from typing import Generic, TypeVar

from rainbowadn.data.collection.collection_interface.collectioninterface import CollectionInterface
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference
from rainbowadn.hashing.rainbow_factory import RainbowFactory

__all__ = ('BlockCollectionInterface',)

BlockType = TypeVar('BlockType')
HeaderType = TypeVar('HeaderType')
ActualStateType = TypeVar('ActualStateType')


class BlockCollectionInterface(
    CollectionInterface[
        BlockType
    ],
    Generic[BlockType, HeaderType, ActualStateType]
):
    def __init__(
            self,
            reference: NullableReference[BlockType]
    ):
        assert isinstance(reference, NullableReference)
        super().__init__(reference)

    def _verify_link(
            self,
            block: BlockType,
            previous: NullableReference[BlockType]
    ) -> bool:
        raise NotImplementedError

    def verify_link(
            self,
            previous: NullableReference[BlockType]
    ) -> bool:
        assert isinstance(previous, NullableReference)
        if self.reference.reference.null():
            return True
        else:
            return self._verify_link(
                self.reference.reference.resolve().resolve(),
                previous
            )

    def _actual_state_factory(self) -> RainbowFactory[ActualStateType]:
        raise NotImplementedError

    def _actual_state(self, block: BlockType) -> HashPoint[ActualStateType]:
        raise NotImplementedError

    def actual_state(self) -> NullableReference[ActualStateType]:
        if self.reference.reference.null():
            return NullableReference(Null(), self._actual_state_factory())
        else:
            return NullableReference.of(self._actual_state(self.reference.reference.resolve().resolve()))

    def add(
            self,
            header: HashPoint[HeaderType]
    ) -> 'BlockCollectionInterface[BlockType, HeaderType, ActualStateType]':
        raise NotImplementedError
